# coding=utf-8
# import all of the modules for writing geometry to dsbXML
from fairyfly.properties import ModelProperties, ShapeProperties, BoundaryProperties
from fairyfly.shape import Shape
from fairyfly.boundary import Boundary
from fairyfly.model import Model
import fairyfly.writer.shape as shape_writer
import fairyfly.writer.boundary as boundary_writer
import fairyfly.writer.model as model_writer

from .properties.model import ModelThermProperties
from .properties.shape import ShapeThermProperties
from .properties.boundary import BoundaryThermProperties
from .writer import model_to_thmz, model_to_therm_xml_str, shape_to_therm_xml_str, \
    boundary_to_therm_xml_str

# set a hidden therm attribute on each core geometry Property class to None
# define methods to produce therm property instances on each Property instance
ModelProperties._therm = None
ShapeProperties._therm = None
BoundaryProperties._therm = None


def model_therm_properties(self):
    if self._therm is None:
        self._therm = ModelThermProperties(self.host)
    return self._therm


def shape_therm_properties(self):
    if self._therm is None:
        self._therm = ShapeThermProperties(self.host)
    return self._therm


def boundary_therm_properties(self):
    if self._therm is None:
        self._therm = BoundaryThermProperties(self.host)
    return self._therm


# add therm property methods to the Properties classes
ModelProperties.therm = property(model_therm_properties)
ShapeProperties.therm = property(shape_therm_properties)
BoundaryProperties.therm = property(boundary_therm_properties)

# add writers to the fairyfly-core modules
shape_writer.therm_xml = shape_to_therm_xml_str
boundary_writer.therm_xml = boundary_to_therm_xml_str
model_writer.therm_xml = model_to_therm_xml_str
model_writer.thmz = model_to_thmz

# add energy writer to core objects
Shape.to_therm_xml = shape_to_therm_xml_str
Boundary.to_therm_xml = boundary_to_therm_xml_str
Model.to_therm_xml = model_to_therm_xml_str
Model.to_thmz = model_to_thmz



# coding=utf-8
"""Solid THERM material."""
from __future__ import division
import xml.etree.ElementTree as ET

from fairyfly._lockable import lockable
from fairyfly.typing import float_in_range, float_positive, uuid_from_therm_id

from ._base import _ThermMaterialBase


@lockable
class SolidMaterial(_ThermMaterialBase):
    """Typical conductive material.

    Args:
        conductivity: Number for the thermal conductivity of the material [W/m-K].
        emissivity: Number between 0 and 1 for the infrared hemispherical
            emissivity of the front side of the material. (Default: 0.9).
        emissivity_back: Number between 0 and 1 for the infrared hemispherical
            emissivity of the back side of the material. If None, this will
            default to the same value specified for emissivity. (Default: None)
        density: Optional number for the density of the material [kg/m3]. (Default: None).
        porosity: Optional number between zero and one for the porosity of
            the material. (Default: None).
        specific_heat: Optional number for the specific heat of the material [J/kg-K].
            If None, it is not included in the export. (Default: None).
        vapor_diffusion_resistance: Optional number for the water vapor diffusion
            resistance factor [Dimensionless]. (Default: None).
        identifier: Text string for a unique object ID. Must be a UUID in the
            format xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx. If None, a UUID will
            automatically be generated. (Default: None).

    Properties:
        * identifier
        * display_name
        * conductivity
        * emissivity
        * emissivity_back
        * density
        * porosity
        * specific_heat
        * vapor_diffusion_resistance
        * resistivity
        * color
        * protected
        * user_data
    """
    __slots__ = ('_conductivity', '_emissivity', '_emissivity_back',
                 '_density', '_porosity', '_specific_heat',
                 '_vapor_diffusion_resistance')

    def __init__(
        self, conductivity, emissivity=0.9, emissivity_back=None,
        density=None, porosity=None, specific_heat=None,
        vapor_diffusion_resistance=None, identifier=None
    ):
        """Initialize therm material."""
        _ThermMaterialBase.__init__(self, identifier)
        self.conductivity = conductivity
        self.emissivity = emissivity
        self.emissivity_back = emissivity_back
        self.density = density
        self.porosity = porosity
        self.specific_heat = specific_heat
        self.vapor_diffusion_resistance = vapor_diffusion_resistance

    @property
    def conductivity(self):
        """Get or set the conductivity of the material layer [W/m-K]."""
        return self._conductivity

    @conductivity.setter
    def conductivity(self, cond):
        self._conductivity = float_positive(cond, 'material conductivity')
        assert self._conductivity != 0, 'Conductivity cannot be equal to zero.'

    @property
    def emissivity(self):
        """Get or set the hemispherical emissivity of the front side of the material."""
        return self._emissivity

    @emissivity.setter
    def emissivity(self, ir_e):
        ir_e = float_in_range(ir_e, 0.0, 1.0, 'material emissivity')
        self._emissivity = ir_e

    @property
    def emissivity_back(self):
        """Get or set the hemispherical emissivity of the back side of the material."""
        return self._emissivity_back if self._emissivity_back is not None \
            else self._emissivity

    @emissivity_back.setter
    def emissivity_back(self, ir_e):
        if ir_e is not None:
            ir_e = float_in_range(ir_e, 0.0, 1.0, 'material emissivity')
        self._emissivity_back = ir_e

    @property
    def density(self):
        """Get or set the density of the material layer [kg/m3]."""
        return self._density

    @density.setter
    def density(self, value):
        if value is not None:
            value = float_positive(value, 'material density')
        self._density = value

    @property
    def porosity(self):
        """Get or set a number between zero and one for the porosity."""
        return self._porosity

    @porosity.setter
    def porosity(self, value):
        if value is not None:
            value = float_in_range(value, 0.0, 1.0, 'material porosity')
        self._porosity = value

    @property
    def specific_heat(self):
        """Get or set the specific heat of the material layer [J/kg-K]."""
        return self._specific_heat

    @specific_heat.setter
    def specific_heat(self, value):
        if value is not None:
            value = float_positive(value, 'material specific heat')
        self._specific_heat = value

    @property
    def vapor_diffusion_resistance(self):
        """Get or set the vapor diffusion resistance of the material."""
        return self._vapor_diffusion_resistance

    @vapor_diffusion_resistance.setter
    def vapor_diffusion_resistance(self, value):
        if value is not None:
            value = float_positive(value, 'vapor diffusion resistance')
        self._vapor_diffusion_resistance = value

    @property
    def resistivity(self):
        """Get or set the resistivity of the material layer [m-K/W]."""
        return 1 / self._conductivity

    @resistivity.setter
    def resistivity(self, resis):
        self._conductivity = 1 / float_positive(resis, 'material resistivity')

    @classmethod
    def from_therm_xml(cls, xml_element):
        """Create a SolidMaterial from an XML element of a THERM Material.

        Args:
            xml_element: An XML element of a THERM material.
        """
        # create the base material from the UUID and conductivity
        xml_uuid = xml_element.find('UUID')
        identifier = xml_uuid.text
        if len(identifier) == 31:
            identifier = uuid_from_therm_id(identifier)
        xml_solid = xml_element.find('Solid')
        xml_hyt = xml_solid.find('HygroThermal')
        xml_cond = xml_hyt.find('ThermalConductivityDry')
        conductivity = xml_cond.text
        mat = SolidMaterial(conductivity, identifier=identifier)
        # assign the other hygrothermal attributes if specified
        xml_dens = xml_hyt.find('BulkDensity')
        if xml_dens is not None:
            mat.density = xml_dens.text
        xml_por = xml_hyt.find('Porosity')
        if xml_por is not None:
            mat.porosity = xml_por.text
        xml_shc = xml_hyt.find('SpecificHeatCapacityDry')
        if xml_shc is not None:
            mat.specific_heat = xml_shc.text
        xml_vdr = xml_hyt.find('WaterVaporDiffusionResistanceFactor')
        if xml_vdr is not None:
            mat.vapor_diffusion_resistance = xml_vdr.text
        # assign the optical properties if specified
        xml_optical = xml_solid.find('Optical')
        if xml_optical is not None:
            xml_int = xml_optical.find('Integrated')
            if xml_int is not None:
                xml_inf = xml_int.find('Infrared')
                if xml_inf is not None:
                    xml_emiss = xml_inf.find('Emissivity-Front')
                    if xml_emiss is not None:
                        mat.emissivity = xml_emiss.text
                    xml_emiss_b = xml_inf.find('Emissivity-Back')
                    if xml_emiss_b is not None:
                        mat.emissivity_back = xml_emiss_b.text
        # assign the name and color if they are specified
        xml_name = xml_element.find('Name')
        if xml_name is not None:
            mat.display_name = xml_name.text
        xml_col = xml_element.find('Color')
        if xml_col is not None:
            mat.color = xml_col.text
        xml_protect = xml_element.find('Protected')
        if xml_protect is not None:
            mat.protected = True if xml_protect.text == 'true' else False
        return mat

    @classmethod
    def from_therm_xml_str(cls, xml_str):
        """Create a SolidMaterial from an XML text string of a THERM Material.

        Args:
            xml_str: An XML text string of a THERM material.
        """
        root = ET.fromstring(xml_str)
        return cls.from_therm_xml(root)

    @classmethod
    def from_dict(cls, data):
        """Create a SolidMaterial from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": 'SolidMaterial',
            "identifier": 'f26f3597-e3ee-43fe-a0b3-ec673f993d86',
            "display_name": 'Concrete',
            "conductivity": 2.31,
            "emissivity": 0.9,
            "emissivity_back": 0.9,
            "density": 2322,
            "porosity": 0.24,
            "specific_heat": 832,
            "vapor_diffusion_resistance": 19
            }
        """
        assert data['type'] == 'SolidMaterial', \
            'Expected SolidMaterial. Got {}.'.format(data['type'])

        emiss = data['emissivity'] if 'emissivity' in data and \
            data['emissivity'] is not None else 0.9
        emiss_b = data['emissivity_back'] if 'emissivity_back' in data else None
        dens = data['density'] if 'density' in data else None
        poro = data['porosity'] if 'porosity' in data else None
        s_heat = data['specific_heat'] if 'specific_heat' in data else None
        vdr = data['vapor_diffusion_resistance'] \
            if 'vapor_diffusion_resistance' in data else None

        new_mat = cls(
            data['conductivity'], emiss, emiss_b, dens, poro, s_heat, vdr,
            data['identifier'])
        if 'display_name' in data and data['display_name'] is not None:
            new_mat.display_name = data['display_name']
        if 'color' in data and data['color'] is not None:
            new_mat.color = data['color']
        if 'protected' in data and data['protected'] is not None:
            new_mat.protected = data['protected']
        if 'user_data' in data and data['user_data'] is not None:
            new_mat.user_data = data['user_data']
        return new_mat

    def to_therm_xml(self, materials_element=None):
        """Get an THERM XML element of the material.

        Args:
            materials_element: An optional XML Element for the Materials to
                which the generated objects will be added. If None, a new XML
                Element will be generated.

        .. code-block:: xml

            <Material>
                <UUID>1543f77f-e4aa-47d4-b7bf-2365d7bc0b9d</UUID>
                <Name>Polyurethane Foam</Name>
                <Protected>false</Protected>
                <Color>0xFFFFC1</Color>
                <Solid>
                    <HygroThermal>
                        <ThermalConductivityDry>0.05</ThermalConductivityDry>
                    </HygroThermal>
                    <Optical>
                        <Integrated>
                            <Infrared>
                                <Emissivity-Front>0.9</Emissivity-Front>
                                <Emissivity-Back>0.9</Emissivity-Back>
                            </Infrared>
                        </Integrated>
                    </Optical>
                </Solid>
            </Material>
        """
        # create a new Materials element if one is not specified
        if materials_element is not None:
            xml_mat = ET.SubElement(materials_element, 'Material')
        else:
            xml_mat = ET.Element('Material')
        # add all of the required basic attributes
        xml_id = ET.SubElement(xml_mat, 'UUID')
        xml_id.text = self.identifier
        xml_name = ET.SubElement(xml_mat, 'Name')
        xml_name.text = self.display_name
        xml_protect = ET.SubElement(xml_mat, 'Protected')
        xml_protect.text = 'true' if self.protected else 'false'
        xml_color = ET.SubElement(xml_mat, 'Color')
        xml_color.text = self.color.to_hex().replace('#', '0x')
        xml_solid = ET.SubElement(xml_mat, 'Solid')
        # add all of the required hygrothermal and optical attributes
        xml_hyt = ET.SubElement(xml_solid, 'HygroThermal')
        xml_cond = ET.SubElement(xml_hyt, 'ThermalConductivityDry')
        xml_cond.text = str(self.conductivity)
        xml_optical = ET.SubElement(xml_solid, 'Optical')
        xml_int = ET.SubElement(xml_optical, 'Integrated')
        xml_inf = ET.SubElement(xml_int, 'Infrared')
        xml_emiss = ET.SubElement(xml_inf, 'Emissivity-Front')
        xml_emiss.text = str(self.emissivity)
        xml_emiss_b = ET.SubElement(xml_inf, 'Emissivity-Back')
        xml_emiss_b.text = str(self.emissivity_back)
        # add any of the optional hygrothermal attributes
        if self.density is not None:
            xml_dens = ET.SubElement(xml_hyt, 'BulkDensity')
            xml_dens.text = str(self.density)
        if self.porosity is not None:
            xml_por = ET.SubElement(xml_hyt, 'Porosity')
            xml_por.text = str(self.porosity)
        if self.specific_heat is not None:
            xml_shc = ET.SubElement(xml_hyt, 'SpecificHeatCapacityDry')
            xml_shc.text = str(self.specific_heat)
        if self.vapor_diffusion_resistance is not None:
            xml_vdr = ET.SubElement(xml_hyt, 'WaterVaporDiffusionResistanceFactor')
            xml_vdr.text = str(self.vapor_diffusion_resistance)
        return xml_mat

    def to_therm_xml_str(self):
        """Get an THERM XML string of the material."""
        xml_root = self.to_therm_xml()
        try:  # try to indent the XML to make it read-able
            ET.indent(xml_root)
            return ET.tostring(xml_root, encoding='unicode')
        except AttributeError:  # we are in Python 2 and no indent is available
            return ET.tostring(xml_root)

    def to_dict(self):
        """SolidMaterial dictionary representation."""
        base = {
            'type': 'SolidMaterial',
            'identifier': self.identifier,
            'conductivity': self.conductivity,
            'emissivity': self.emissivity
        }
        if self._emissivity_back is not None:
            base['emissivity_back'] = self._emissivity_back
        if self._density is not None:
            base['density'] = self._density
        if self._porosity is not None:
            base['porosity'] = self.porosity
        if self._specific_heat is not None:
            base['specific_heat'] = self.specific_heat
        if self._vapor_diffusion_resistance is not None:
            base['vapor_diffusion_resistance'] = self.vapor_diffusion_resistance
        if self._display_name is not None:
            base['display_name'] = self.display_name
        base['protected'] = self._protected
        base['color'] = self.color.to_hex()
        if self._user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.identifier, self.conductivity, self.emissivity,
                self.emissivity_back, self.density, self.porosity,
                self.specific_heat, self.vapor_diffusion_resistance)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, SolidMaterial) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __copy__(self):
        new_material = self.__class__(
            self.conductivity, self._emissivity, self._emissivity_back,
            self._density, self._porosity, self._specific_heat,
            self._vapor_diffusion_resistance, self.identifier)
        new_material._display_name = self._display_name
        new_material._color = self._color
        new_material._protected = self._protected
        new_material._user_data = None if self._user_data is None \
            else self._user_data.copy()
        return new_material

    def __repr__(self):
        return 'Solid THERM Material: {}'.format(self.display_name)



"""Load all conditions from the LBNL XML files and JSON libraries."""
import os
import json

from fairyfly_therm.config import folders
from fairyfly_therm.condition.steadystate import SteadyState

# empty dictionaries to hold loaded conditions
_conditions = {}
_FUNDAMENTALS = ('Adiabatic', 'Frame Cavity Surface',
                 'Generic Exterior', 'Generic Interior')

# ensure that there is always an adiabatic and frame cavity condition
adiabatic_dict = {
    'type': 'SteadyState',
    'identifier': '61d7bd1c-22c6-4ea0-8720-0696e8c194ad',
    'temperature': 0,
    'film_coefficient': 0,
    'display_name': 'Adiabatic',
    'protected': True,
    'color': '#000000'
}
adiabatic = SteadyState.from_dict(adiabatic_dict)
adiabatic.lock()
_conditions[adiabatic.display_name] = adiabatic

frame_cavity_dict = {
    'type': 'SteadyState',
    'identifier': '320f2027-a8f4-44d3-907f-1d90f5118632',
    'temperature': 0,
    'film_coefficient': 0,
    'display_name': 'Frame Cavity Surface',
    'protected': True,
    'color': '#FF0000'
}
frame_cavity = SteadyState.from_dict(frame_cavity_dict)
frame_cavity.lock()
_conditions[frame_cavity.display_name] = frame_cavity

# add generic indoor and outdoor conditions
exterior_dict = {
    'type': 'SteadyState',
    'identifier': 'de5cfad3-8a58-48b7-8583-b31a2650bc80',
    'temperature': -18.0,
    'film_coefficient': 26.0,
    'display_name': 'Generic Exterior',
    'protected': True,
    'color': '#0080c0'
}
exterior = SteadyState.from_dict(exterior_dict)
exterior.lock()
_conditions[exterior.display_name] = exterior

interior_dict = {
    'type': 'SteadyState',
    'identifier': '7eba559e-ebbc-4dc4-82a7-266a888fe5a5',
    'temperature': 21.0,
    'film_coefficient': 3.12,
    'display_name': 'Generic Interior',
    'protected': True,
    'color': '#ff8040'
}
interior = SteadyState.from_dict(interior_dict)
interior.lock()
_conditions[interior.display_name] = interior


# load the conditions from the LBNL library if they exist
if folders.bc_steady_state_lib_file is not None:
    conds = SteadyState.extract_all_from_xml_file(folders.bc_steady_state_lib_file)
    for con in conds:
        con.lock()
        _conditions[con.display_name] = con


def check_and_add_condition(con):
    """Check that a condition is not overwriting a default and add it."""
    con.lock()
    if con.display_name not in _FUNDAMENTALS:
        if isinstance(con, SteadyState):
            _conditions[con.display_name] = con


def load_conditions_from_folder(lib_folder):
    """Load all of the condition objects from a therm standards folder.

    Args:
        lib_folder: Path to a sub-folder within a honeybee standards folder.
    """
    for f in os.listdir(lib_folder):
        f_path = os.path.join(lib_folder, f)
        if os.path.isfile(f_path):
            if f_path.endswith('.xml'):
                conds = SteadyState.extract_all_from_xml_file(f_path)
                for c in conds:
                    check_and_add_condition(c)
            elif f_path.endswith('.json'):
                with open(f_path) as json_file:
                    data = json.load(json_file)
                if 'type' in data:  # single object
                    if data['type'] == 'SteadyState':
                        check_and_add_condition(SteadyState.from_dict(data))
                else:  # a collection of several objects
                    for m_id in data:
                        try:
                            m_dict = data[m_id]
                            if m_dict['type'] == 'SteadyState':
                                check_and_add_condition(
                                    SteadyState.from_dict(m_dict))
                        except (TypeError, KeyError):
                            pass  # not an acceptable JSON; possibly a comment


# load therm gases from a user folder if we are not using the official THERM lib
if folders.therm_lib_path is not None:
    load_conditions_from_folder(folders.therm_lib_path)



"""Fairyfly therm simulation running commands."""
import click
import sys
import logging

from ladybug.commandutil import process_content_to_output
from fairyfly.model import Model
from fairyfly_therm.run import run_model

_logger = logging.getLogger(__name__)


@click.group(help='Commands for simulating Fairyfly files in THERM.')
def simulate():
    pass


@simulate.command('model')
@click.argument('model-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option(
    '--folder', '-f', help='Folder on this computer, into which the THMZ '
    'and simulation files will be written. If None, the files will be output '
    'to the fairyfly default simulation folder and placed in a project '
    'folder with the same name as the model-file.',
    default=None, show_default=True,
    type=click.Path(file_okay=False, dir_okay=True, resolve_path=True))
@click.option(
    '--log-file', '-log', help='Optional log file to output the paths of the '
    'generated THMZ file if successfully created. By default the list will '
    'be printed out to stdout', type=click.File('w'), default='-', show_default=True)
def simulate_model_cli(model_file, folder, log_file):
    """Simulate a Fairyfly Model in THERM.

    \b
    Args:
        model_file: Full path to a Model file as either a FFJSON or FFpkl.
    """
    try:
        simulate_model(model_file, folder, log_file)
    except Exception as e:
        _logger.exception('Model simulation failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def simulate_model(model_file, folder=None, log_file=None):
    """Simulate a Fairyfly Model in THERM.

    Args:
        model_file: Full path to a Fairyfly Model file (FFJSON or FFpkl).
        Folder on this computer, into which the THMZ and simulation files will
            be written. If None, the files will be output to the fairyfly
            default simulation folder and placed in a project folder with the
            same name as the model-file.
        output_file: Optional THMZ file path to output the THMZ string of the
            translation. If None, the string will be returned from this function.
    """
    model = Model.from_file(model_file)
    thmz_file = run_model(model, folder, silent=True)
    process_content_to_output(thmz_file, log_file)



# coding=utf-8
"""Cavity THERM material."""
from __future__ import division
import xml.etree.ElementTree as ET

from fairyfly._lockable import lockable
from fairyfly.typing import float_in_range, therm_id_from_uuid, uuid_from_therm_id

from ._base import _ThermMaterialBase
from .gas import Gas
from ..lib.gases import air


@lockable
class CavityMaterial(_ThermMaterialBase):
    """Typical cavity material.

    Args:
        gas: A Gas material object for the gas that fills the cavity. (Default: air).
        cavity_model: Text for the type of cavity model to be used to determine
            the thermal resistance of the material. Choose from the following:

            * CEN
            * NFRC
            * ISO15099
            * ISO15099Ventilated

        emissivity: Number between 0 and 1 for the infrared hemispherical
            emissivity of the front side of the material. (Default: 0.9).
        emissivity_back: Number between 0 and 1 for the infrared hemispherical
            emissivity of the back side of the material. If None, this will
            default to the same value specified for emissivity. (Default: None)
        identifier: Text string for a unique object ID. Must be a UUID in the
            format xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx. If None, a UUID will
            automatically be generated. (Default: None).

    Properties:
        * identifier
        * display_name
        * therm_uuid
        * gas
        * cavity_model
        * emissivity
        * emissivity_back
        * color
        * protected
        * user_data
    """
    __slots__ = ('_gas', '_cavity_model', '_emissivity', '_emissivity_back')
    CAVITY_MODELS = ('CEN', 'NFRC', 'ISO15099', 'ISO15099Ventilated')

    def __init__(
        self, gas=air, cavity_model='CEN', emissivity=0.9, emissivity_back=None,
        identifier=None
    ):
        """Initialize therm material."""
        _ThermMaterialBase.__init__(self, identifier)
        self.gas = gas
        self.cavity_model = cavity_model
        self.emissivity = emissivity
        self.emissivity_back = emissivity_back

    @property
    def gas(self):
        """Get or set a Gas object used to denote the gas in the cavity."""
        return self._gas

    @gas.setter
    def gas(self, value):
        if value is not None:
            assert isinstance(value, Gas), 'Expected Gas object for CavityMaterial ' \
                'gas. Got {}.'.format(type(value))
        else:
            value = air
        self._gas = value

    @property
    def cavity_model(self):
        """Get or set text for the convection model to be used in the cavity."""
        return self._cavity_model

    @cavity_model.setter
    def cavity_model(self, value):
        if value is not None:
            clean_input = str(value).lower()
            for key in self.CAVITY_MODELS:
                if key.lower() == clean_input:
                    value = key
                    break
            else:
                raise ValueError(
                    'Material cavity_model "{}" is not supported.\n'
                    'Choose from the following:\n{}'.format(
                        value, '\n'.join(self.CAVITY_MODELS)))
            self._cavity_model = value
        else:
            self._cavity_model = self.CAVITY_MODELS[0]

    @property
    def emissivity(self):
        """Get or set the hemispherical emissivity of the front side of the material."""
        return self._emissivity

    @emissivity.setter
    def emissivity(self, ir_e):
        ir_e = float_in_range(ir_e, 0.0, 1.0, 'material emissivity')
        self._emissivity = ir_e

    @property
    def emissivity_back(self):
        """Get or set the hemispherical emissivity of the back side of the material."""
        return self._emissivity_back if self._emissivity_back is not None \
            else self._emissivity

    @emissivity_back.setter
    def emissivity_back(self, ir_e):
        if ir_e is not None:
            ir_e = float_in_range(ir_e, 0.0, 1.0, 'material emissivity')
        self._emissivity_back = ir_e

    @property
    def therm_uuid(self):
        """Get the UUID of this object as it would appear in a THERM XML or thmz file.

        This is always derived from the object identifier but this is slightly
        different than standard UUIDs, which have 4 more values in a 8-4-4-4-12
        structure instead of a 8-4-4-12 structure used by THERM.
        """
        return therm_id_from_uuid(self._identifier)

    @classmethod
    def from_therm_xml(cls, xml_element, gases):
        """Create a CavityMaterial from an XML element of a THERM Material.

        Args:
            xml_element: An XML element of a THERM material.
            gases: A dictionary with gas names as keys and Gas object instances
                as values. These will be used to reassign the gas that fills
                this cavity.
        """
        # create the base material from the UUID and conductivity
        xml_uuid = xml_element.find('UUID')
        identifier = xml_uuid.text
        if len(identifier) == 31:
            identifier = uuid_from_therm_id(identifier)
        xml_cavity = xml_element.find('Cavity')
        xml_c_model = xml_cavity.find('CavityStandard')
        cavity_model = xml_c_model.text
        xml_gas = xml_cavity.find('Gas')
        try:
            gas = gases[xml_gas.text]
        except KeyError as e:
            raise ValueError('Failed to find {} in gases.'.format(e))
        mat = CavityMaterial(gas, cavity_model, identifier=identifier)
        # assign the other attributes if specified
        xml_emiss1 = xml_cavity.find('EmissivitySide1')
        if xml_emiss1 is not None:
            mat.emissivity = xml_emiss1.text
        xml_emiss2 = xml_cavity.find('EmissivitySide2')
        if xml_emiss2 is not None:
            mat.emissivity_back = xml_emiss2.text
        # assign the name and color if they are specified
        xml_name = xml_element.find('Name')
        if xml_name is not None:
            mat.display_name = xml_name.text
        xml_col = xml_element.find('Color')
        if xml_col is not None:
            mat.color = xml_col.text
        xml_protect = xml_element.find('Protected')
        if xml_protect is not None:
            mat.protected = True if xml_protect.text == 'true' else False
        return mat

    @classmethod
    def from_therm_xml_str(cls, xml_str, gases):
        """Create a CavityMaterial from an XML text string of a THERM Material.

        Args:
            xml_str: An XML text string of a THERM material.
            gases: A dictionary with gas names as keys and Gas object instances
                as values. These will be used to reassign the gas that fills
                this cavity.
        """
        root = ET.fromstring(xml_str)
        return cls.from_therm_xml(root, gases)

    @classmethod
    def from_dict(cls, data):
        """Create a CavityMaterial from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": 'CavityMaterial',
            "identifier": '0b46bbd7-0dbc-c148-3afe87431bf0',
            "display_name": 'Frame Cavity - CEN Simplified',
            "gas": {},  # dictionary definition of a gas
            "cavity_model": "CEN",
            "emissivity": 0.9,
            "emissivity_back": 0.9
            }
        """
        assert data['type'] == 'CavityMaterial', \
            'Expected CavityMaterial. Got {}.'.format(data['type'])

        emiss = data['emissivity'] if 'emissivity' in data and \
            data['emissivity'] is not None else 0.9
        emiss_b = data['emissivity_back'] if 'emissivity_back' in data else None
        new_mat = cls(
            Gas.from_dict(data['gas']), data['cavity_model'], emiss, emiss_b,
            data['identifier'])
        cls._assign_optional_from_dict(new_mat, data)
        return new_mat

    @classmethod
    def from_dict_abridged(cls, data, gases):
        """Create a Gas from an abridged dictionary.

        Args:
            data: An GasAbridged dictionary.
            gases: A dictionary with Gas identifiers as keys and Gas object instances
                as values. These will be used to reassign the gas that fills
                this cavity.

        .. code-block:: python

            {
            "type": 'CavityMaterial',
            "identifier": '0b46bbd7-0dbc-c148-3afe87431bf0',
            "display_name": 'Frame Cavity - CEN Simplified',
            "gas": '6c2409e9-5296-46c1-be11-9029b59a549b',
            "cavity_model": "CEN",
            "emissivity": 0.9,
            "emissivity_back": 0.9
            }
        """
        assert data['type'] == 'CavityMaterialAbridged', \
            'Expected CavityMaterialAbridged. Got {}.'.format(data['type'])
        try:
            gas_obj = gases[data['gas']]
        except KeyError as e:
            raise ValueError('Failed to find {} in gases.'.format(e))
        emiss = data['emissivity'] if 'emissivity' in data and \
            data['emissivity'] is not None else 0.9
        emiss_b = data['emissivity_back'] if 'emissivity_back' in data else None
        new_mat = cls(gas_obj, data['cavity_model'], emiss, emiss_b, data['identifier'])
        cls._assign_optional_from_dict(new_mat, data)
        return new_mat

    @staticmethod
    def _assign_optional_from_dict(new_obj, data):
        """Assign optional attributes when serializing from dict."""
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'color' in data and data['color'] is not None:
            new_obj.color = data['color']
        if 'protected' in data and data['protected'] is not None:
            new_obj.protected = data['protected']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']

    def to_therm_xml(self, materials_element=None):
        """Get an THERM XML element of the material.

        Args:
            materials_element: An optional XML Element for the Materials to
                which the generated objects will be added. If None, a new XML
                Element will be generated.

        .. code-block:: xml

            <Material>
                <UUID>0b46bbd7-0dbc-c148-3afe87431bf0</UUID>
                <Name>Frame Cavity - CEN Simplified</Name>
                <Protected>false</Protected>
                <Color>0xB3FFB3</Color>
                <Cavity>
                    <CavityStandard>CEN</CavityStandard>
                    <Gas>Air</Gas>
                    <EmissivitySide1>0.9</EmissivitySide1>
                    <EmissivitySide2>0.9</EmissivitySide2>
                </Cavity>
            </Material>
        """
        # create a new Materials element if one is not specified
        if materials_element is not None:
            xml_mat = ET.SubElement(materials_element, 'Material')
        else:
            xml_mat = ET.Element('Material')
        # add all of the required basic attributes
        xml_id = ET.SubElement(xml_mat, 'UUID')
        xml_id.text = self.identifier
        xml_name = ET.SubElement(xml_mat, 'Name')
        xml_name.text = self.display_name
        xml_protect = ET.SubElement(xml_mat, 'Protected')
        xml_protect.text = 'true' if self.protected else 'false'
        xml_color = ET.SubElement(xml_mat, 'Color')
        xml_color.text = self.color.to_hex().replace('#', '0x')
        xml_cavity = ET.SubElement(xml_mat, 'Cavity')
        # add all of the required cavity attributes
        xml_model = ET.SubElement(xml_cavity, 'CavityStandard')
        xml_model.text = self.cavity_model
        xml_gas = ET.SubElement(xml_cavity, 'Gas')
        xml_gas.text = self.gas.display_name
        xml_emiss = ET.SubElement(xml_cavity, 'EmissivitySide1')
        xml_emiss.text = str(self.emissivity)
        xml_emiss_b = ET.SubElement(xml_cavity, 'EmissivitySide2')
        xml_emiss_b.text = str(self.emissivity_back)
        return xml_mat

    def to_therm_xml_str(self):
        """Get an THERM XML string of the material."""
        xml_root = self.to_therm_xml()
        try:  # try to indent the XML to make it read-able
            ET.indent(xml_root)
            return ET.tostring(xml_root, encoding='unicode')
        except AttributeError:  # we are in Python 2 and no indent is available
            return ET.tostring(xml_root)

    def to_dict(self, abridged=False):
        """CavityMaterial dictionary representation."""
        base = {'type': 'CavityMaterial'} if not abridged \
            else {'type': 'CavityMaterialAbridged'}
        base['identifier'] = self.identifier
        base['gas'] = self.gas.identifier if abridged else self.gas.to_dict()
        base['cavity_model'] = self.cavity_model
        base['emissivity'] = self.emissivity
        if self._emissivity_back is not None:
            base['emissivity_back'] = self.emissivity_back
        if self._display_name is not None:
            base['display_name'] = self.display_name
        base['protected'] = self._protected
        base['color'] = self.color.to_hex()
        if self._user_data is not None:
            base['user_data'] = self.user_data
        return base

    def lock(self):
        """The lock() method will also lock the gas."""
        self._locked = True
        self.gas.lock()

    def unlock(self):
        """The unlock() method will also unlock the gas."""
        self._locked = False
        self.gas.unlock()

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.therm_uuid, hash(self.gas), self.cavity_model,
                self.emissivity, self.emissivity_back)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, CavityMaterial) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __copy__(self):
        new_material = self.__class__(
            self.gas.duplicate(), self.cavity_model,
            self._emissivity, self._emissivity_back, self.identifier)
        new_material._display_name = self._display_name
        new_material._color = self._color
        new_material._protected = self._protected
        new_material._user_data = None if self._user_data is None \
            else self._user_data.copy()
        return new_material

    def __repr__(self):
        return 'Cavity THERM Material: {}'.format(self.display_name)



# coding=utf-8
"""Module for running files through THERM CLI."""
from __future__ import division

import os
import subprocess

from ladybug.futil import write_to_file, preparedir
from fairyfly.typing import clean_string
from fairyfly.config import folders as ff_folders
from fairyfly.model import Model
from .config import folders


def run_model(model, directory=None, silent=False):
    """Run a fairyfly Model through THERM CLI.

    Args:
        model: Path to a THMZ file to be run using THERM CLI.
        directory: The directory in which the simulation files will be written.
            If None, this will default to the fairyfly-core default_simulation_folder.
        silent: Boolean to note whether the THERM simulation should be run silently.

    Returns:
        The path to the input thmz_file with results inside of it.
    """
    assert isinstance(model, Model), \
        'Expected Fairyfly Model. Got {}.'.format(type(model))
    # get a default directory if none was specified
    if directory is None:
        model_name = clean_string(model.display_name)
        directory = os.path.join(ff_folders.default_simulation_folder, model_name)
    preparedir(directory)
    # write the Model to a .thmz file
    thmz_file = os.path.join(directory, 'model.thmz')
    model.to_thmz(thmz_file)
    # run the thmz_file through THERM
    thmz_file = run_thmz(thmz_file, silent)
    # parse the log file to check if there were any failures
    log_file = os.path.join(directory, 'therm.log')
    with open(log_file, 'r') as lf:
        sim_log = lf.read()
    if 'Calculation complete.' not in sim_log:
        msg = 'THERM simulation failed. Open the thmz file in the THERM interface ' \
            'for more info.\n{}'.format(sim_log)
        raise ValueError(msg)
    return thmz_file


def run_thmz(thmz_file, silent=False):
    """Run a .thmz file using the THERM CLI.

    Args:
        thmz_file: Path to a THMZ file to be run using THERM CLI.
        silent: Boolean to note whether the THERM simulation should be run silently.

    Returns:
        The path to the input thmz_file with results inside of it.
    """
    # check that THERM is installed on the machine
    assert folders.therm_exe is not None, \
        'No usable THERM executable was found on the machine.'
    # check that the THMZ file exists
    thmz_file = os.path.abspath(thmz_file)
    assert os.path.isfile(thmz_file), 'No THMZ file found at {}.'.format(thmz_file)
    directory = os.path.split(thmz_file)[0]
    log_file = os.path.join(directory, 'therm.log')

    # write a batch file to call THERM CLI; useful for manually re-running the sim
    if not silent:
        working_drive = directory[:2]
        batch = '{}\n"{}" -pw thmCLA -thmz "{}" -log "{}" -calc -exit'.format(
            working_drive, folders.therm_exe, thmz_file, log_file)
        if all(ord(c) < 128 for c in batch):  # just run the batch file as it is
            batch_file = os.path.join(directory, 'run_therm.bat')
            write_to_file(batch_file, batch, True)
            os.system('"{}"'.format(batch_file))  # run the batch file
            return thmz_file

    # given .bat file restrictions with non-ASCII characters, run the sim with subprocess
    cmds = [folders.therm_exe, '-pw', 'thmCLA', '-thmz', thmz_file,
            '-log', log_file, '-calc', '-exit']
    process = subprocess.Popen(cmds, shell=silent)
    process.communicate()  # prevents the script from running before command is done

    return thmz_file



# coding=utf-8
"""Module for parsing THERM results."""
from __future__ import division

import os
import zipfile
import json
import re
import xml.etree.ElementTree as ET

from ladybug_geometry.geometry2d import Point2D
from ladybug_geometry.geometry3d import Vector3D, Plane, Mesh3D


class THMZResult(object):
    """Object for parsing results out of simulated THMZ files.

    Args:
        file_path: Full path to a THMZ file that was simulated using THERM.

    Properties:
        * file_path
        * u_factors
        * plane
        * mesh
        * temperatures
        * heat_fluxes
        * heat_flux_magnitudes
    """

    def __init__(self, file_path):
        """Initialize THMZResult."""
        assert os.path.isfile(file_path), 'No file was found at {}'.format(file_path)
        assert file_path.lower().endswith('.thmz'), \
            '"{}" is not a THMZ file ending in .thmz.'.format(file_path)
        self._file_path = file_path

        # set up variables to track what has been loaded from the file
        self._plane_loaded = False
        self._mesh_loaded = False
        self._u_factors_loaded = False
        self._ss_mesh_results_loaded = False

        # values to be computed as soon as they are requested
        self._plane = None
        self._u_factors = None
        self._mesh = None
        self._temperatures = None
        self._heat_fluxes = None
        self._heat_flux_magnitudes = None

    @property
    def file_path(self):
        """Get the path to the .thmz file."""
        return self._file_path

    @property
    def u_factors(self):
        """Get a tuple of UFactor objects for the .

        This will be None if there is no U-Factor information in the THMZ file
        and this will be an empty tuple if the model had no U-Factor tags assigned
        to it.
        """
        if not self._u_factors_loaded:
            self._extract_u_factors()
        return self._u_factors

    @property
    def plane(self):
        """Get a ladybug-geometry Plane for the 3D plane in which the mesh exists.

        This will be the World XY plane if there is no plane information in the
        THMZ file.
        """
        if not self._plane_loaded:
            self._extract_plane()
        return self._plane

    @property
    def mesh(self):
        """Get a ladybug-geometry Mesh3D for the finite element mesh of the model.

        Will be None if the THMZ file has not been simulated and there is no Mesh
        in the file.
        """
        if not self._mesh_loaded:
            self._extract_mesh()
        return self._mesh

    @property
    def temperatures(self):
        """Get a tuple of temperatures in Celsius that correspond to the mesh vertices.

        Will be None if the THMZ file has not been simulated and there are no
        steady state results in the file.
        """
        if not self._ss_mesh_results_loaded:
            self._extract_steady_state_results()
        return self._temperatures

    @property
    def heat_fluxes(self):
        """Get a tuple of Vector3Ds that correspond to the mesh vertices for heat fluxes.

        Will be None if the THMZ file has not been simulated and there are no
        steady state results in the file.
        """
        if not self._ss_mesh_results_loaded:
            self._extract_steady_state_results()
        return self._heat_fluxes

    @property
    def heat_flux_magnitudes(self):
        """Get a tuple of heat flux values in W/m2 in that correspond to mesh vertices.

        Will be None if the THMZ file has not been simulated and there are no
        steady state results in the file.
        """
        if not self._ss_mesh_results_loaded:
            self._extract_steady_state_results()
        return self._heat_flux_magnitudes

    def _extract_u_factors(self):
        """Extract U-factor results from the THMZ file."""
        self._u_factors_loaded = True
        try:  # load the root of the SteadyStateResults.xml
            with zipfile.ZipFile(self.file_path, 'r') as archive:
                with archive.open('SteadyStateResults.xml') as f:
                    # Read content as bytes and decode to a string
                    content = f.read().decode('utf-8')
        except KeyError:  # no results in the file; it has not been simulated
            return
        # extract the temperatures and heat fluxes from the model
        xml_root = ET.fromstring(content)
        xml_case = xml_root.find('Case')
        u_factors = []
        for xml_u_fac in xml_case:
            if xml_u_fac.tag == 'U-factors':
                u_factors.append(UFactor(xml_u_fac))
        self._u_factors = tuple(u_factors)

    def _extract_plane(self):
        """Extract a Plane object from the THMZ file."""
        self._plane_loaded = True
        # load the root of the Model.xml
        with zipfile.ZipFile(self.file_path, 'r') as archive:
            with archive.open('Model.xml') as f:
                # Read content as bytes and decode to a string
                content = f.read().decode('utf-8')
        xml_root = ET.fromstring(content)
        # extract the Plane specification form the model
        xml_props = xml_root.find('Properties')
        xml_gen = xml_props.find('General')
        xml_notes = xml_gen.find('Notes')
        all_notes = xml_notes.text
        _plane_pattern = re.compile(r"Plane:\s(.*)")
        matches = _plane_pattern.findall(all_notes)
        if len(matches) > 0:
            self._plane = Plane.from_dict(json.loads(matches[0]))
        else:
            self._plane = Plane()

    def _extract_mesh(self):
        """Extract a Mesh3D object from the THMZ file."""
        self._mesh_loaded = True
        try:  # load the root of the Mesh.xml
            with zipfile.ZipFile(self.file_path, 'r') as archive:
                with archive.open('Mesh.xml') as f:
                    # Read content as bytes and decode to a string
                    content = f.read().decode('utf-8')
        except KeyError:  # no mesh in the file; it has not been simulated
            return
        # get the mesh information from the root
        xml_root = ET.fromstring(content)
        xml_case = xml_root.find('Case')
        # extract the vertices (aka. nodes) from the model
        plane = self.plane
        vertices = []
        for xml_node in xml_case.find('Nodes'):
            pt_2d = Point2D(xml_node.find('x').text, xml_node.find('y').text)
            vertices.append(plane.xy_to_xyz(pt_2d))
        # extract the faces (aka. elements) from the model
        faces = []
        for xml_face in xml_case.find('Elements'):
            face_i = []
            for e_prop in xml_face:
                if e_prop.tag.startswith('node'):
                    face_i.append(int(e_prop.text) - 1)
            faces.append(tuple(face_i))
        self._mesh = Mesh3D(vertices, faces)

    def _extract_steady_state_results(self):
        """Extract steady state results from the THMZ file."""
        self._mesh_loaded = True
        try:  # load the root of the SteadyStateMeshResults.xml
            with zipfile.ZipFile(self.file_path, 'r') as archive:
                with archive.open('SteadyStateMeshResults.xml') as f:
                    # Read content as bytes and decode to a string
                    content = f.read().decode('utf-8')
        except KeyError:  # no results in the file; it has not been simulated
            return
        # extract the temperatures and heat fluxes from the model
        xml_root = ET.fromstring(content)
        xml_case = xml_root.find('Case')
        plane = self.plane
        temperatures, heat_fluxes, flux_magnitudes = [], [], []
        for xml_node in xml_case.find('Nodes'):
            temperatures.append(float(xml_node.find('Temperature').text))
            pt_2d = Point2D(xml_node.find('X-flux').text, xml_node.find('Y-flux').text)
            pt_3d = plane.xy_to_xyz(pt_2d)
            flux_vec = Vector3D(pt_3d.x, pt_3d.y, pt_3d.z)
            heat_fluxes.append(flux_vec)
            flux_magnitudes.append(flux_vec.magnitude)
        self._temperatures = tuple(temperatures)
        self._heat_fluxes = tuple(heat_fluxes)
        self._heat_flux_magnitudes = tuple(flux_magnitudes)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return 'THMZ Result: {}'.format(self.file_path)


class UFactor(object):
    """Object for holding the results of an individual U-factor tag.

    Args:
        xml_element: An XML element for a U-factor result in the
            SteadyStateResults.xml file.

    Properties:
        * name
        * delta_temperature
        * heat_flux
        * total_u_factor
        * total_length
        * projected_x_u_factor
        * projected_x_length
        * projected_y_u_factor
        * projected_y_length
        * projected_in_glass_plane_u_factor
        * projected_in_glass_plane_length
        * custom_rotation_u_factor
        * custom_rotation_length
    """

    __slots__ = (
        '_name', '_delta_temperature', '_heat_flux', '_total_u_factor', '_total_length',
        '_projected_x_u_factor', '_projected_x_length',
        '_projected_y_u_factor', '_projected_y_length',
        '_projected_in_glass_plane_u_factor', '_projected_in_glass_plane_length',
        '_custom_rotation_u_factor', '_custom_rotation_length')

    def __init__(self, xml_element):
        """Initialize UFactor."""
        # get the basic properties like the name and heat flux
        self._name = xml_element.find('Tag').text
        self._delta_temperature = float(xml_element.find('DeltaT').text)
        self._heat_flux = float(xml_element.find('HeatFlux').text)

        # set defaults for the different U-factors in case they are not found
        self._total_u_factor = None
        self._total_length = 0
        self._projected_x_u_factor = None
        self._projected_x_length = 0
        self._projected_y_u_factor = None
        self._projected_y_length = 0
        self._projected_in_glass_plane_u_factor = None
        self._projected_in_glass_plane_length = 0
        self._custom_rotation_u_factor = None
        self._custom_rotation_length = 0

        # extract the different types of U-factors
        for xml_project in xml_element:
            if xml_project.tag == 'Projection':
                len_type = xml_project.find('Length-type').text
                if len_type == 'Total Length':
                    self._set_u_factors(
                        xml_project, '_total_u_factor', '_total_length')
                elif len_type == 'Projected X':
                    self._set_u_factors(
                        xml_project, '_projected_x_u_factor', '_projected_x_length')
                elif len_type == 'Projected Y':
                    self._set_u_factors(
                        xml_project, '_projected_y_u_factor', '_projected_y_length')
                elif len_type == 'Projected in glass plane':
                    self._set_u_factors(
                        xml_project, '_projected_in_glass_plane_u_factor',
                        '_projected_in_glass_plane_length')
                elif len_type == 'Custom Rotation':
                    self._set_u_factors(
                        xml_project, '_custom_rotation_u_factor',
                        '_custom_rotation_length')

    @property
    def name(self):
        """Get the name of the U-factor tag."""
        return self._name

    @property
    def delta_temperature(self):
        """Get a number for the temperature delta across the boundaries in Celsius."""
        return self._delta_temperature

    @property
    def heat_flux(self):
        """Get a number for the heat flux across the boundaries in W/m2."""
        return self._heat_flux

    @property
    def total_u_factor(self):
        """Get a number for the total U-Factor across the boundaries in W/m2-K."""
        return self._total_u_factor

    @property
    def total_length(self):
        """Get a number for the total length the boundaries in mm."""
        return self._total_length

    @property
    def projected_x_u_factor(self):
        """Get a number for the X-projected U-Factor across the boundaries in W/m2-K."""
        return self._projected_x_u_factor

    @property
    def projected_x_length(self):
        """Get a number for the X-projected length the boundaries in mm."""
        return self._projected_x_length

    @property
    def projected_y_u_factor(self):
        """Get a number for the X-projected U-Factor across the boundaries in W/m2-K."""
        return self._projected_y_u_factor

    @property
    def projected_y_length(self):
        """Get a number for the Y-projected length the boundaries in mm."""
        return self._projected_y_length

    @property
    def projected_in_glass_plane_u_factor(self):
        """Get a number for the glass plane-projected U-Factor in W/m2-K."""
        return self._projected_in_glass_plane_u_factor

    @property
    def projected_in_glass_plane_length(self):
        """Get a number for the glass plane-projected length in mm."""
        return self._projected_in_glass_plane_length

    @property
    def custom_rotation_u_factor(self):
        """Get a number for the custom rotation-projected U-Factor in W/m2-K."""
        return self._custom_rotation_u_factor

    @property
    def custom_rotation_length(self):
        """Get a number for the custom rotation-projected length in mm."""
        return self._custom_rotation_length

    def _set_u_factors(self, xml_project, u_fac_attr, len_attr):
        """Set the U-Factor properties given an XML Projection element."""
        xml_u_fac = xml_project.find('U-factor')
        if xml_u_fac is not None:
            setattr(self, u_fac_attr, float(xml_u_fac.text))
        xml_len = xml_project.find('Length')
        if xml_len is not None:
            setattr(self, len_attr, float(xml_len.text))

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return 'UFactor: {}'.format(self.name)
